{
    
    xml2gui - A unit to parse an XML file and automatically build a
    GUI using Hag.
    
    ---------------------------------------
    
    Copyright (c) 2010 Ben Golightly

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
    
    --------------------------------------- 

}
Unit
    Uses
        hagC2D, // NOT included, available from tophatstuff.co.uk (+source, MIT licensed)
        poxml   // included, available from tophatstuff.co.uk (+source, MIT licensed)
        
// For hagC2D 20100914 or later

Const
    X2G_RENDER_AFTER_EVERY_LOADED_ELEMENT = FALSE // allows you to see the gui being built. For testing/debugging.

Function SafeXmlParameter(node: ^poxNodes, param: String) : String
Var
    index: Integer
Begin
    index = PoxParameterIndexByName(node, param)
    If index > -1 then
        result = PoxParameter(node, index)
    Else
        result = ""
    Endif
End

Procedure Xml2Gui_LoadForm(parentForm: ^forms, root: ^poxNodes, prefix: String)
Var
    i, j, k, c, c2: Integer
    s: String
    child: ^poxNodes
    tabChild: ^poxNodes
    optionChild: ^poxNodes
    
    // tmp
    childForm: ^forms
    guiButton: ^buttons
    guiCanvas: ^canvases
    guiCheckBox: ^checkBoxes
    guiDropDown: ^dropDowns
    guiFrame: ^frames
    guiListBox: ^listBoxes
    guiProgressBar: ^progressBars
    guiSlider: ^sliders
    guiTabGroup: ^tabGroups
    guiTab: ^tabs
    guiText: ^hTexts
    guiTextBox: ^textBoxes
    tmpImg: Element
    
    // params
    hook: String
    x, y, w, h, l, d, o, f, p, styleId: Integer
    selected, checked: Boolean
    txt, title, toolTip, style: String
    fn: String
    
Begin

    // Init the form
    parentForm.hook = prefix    
    
    // Child elements    
    c = PoxChildren(root)
   
    For i = 0 to (c-1)
        
        If X2G_RENDER_AFTER_EVERY_LOADED_ELEMENT then
            HagUpdateOnce()
            Flip
        Endif    
    
        child = PoxChild(root, i)
        s = child.name.s
        hook = prefix+"."+SafeXmlParameter(child, 'id')
        
        Conditions
        
            // Form
            (s = "form"):
                
                // <form name="Sidebar" id="form123">
            
                Xml2Gui_error("Nested child forms are currently unsupported.", fn)
                
                // Due to a possible bug in Cobra to do with recursive types, nested child forms
                // won't work! 
            
            
            // Button
            (s = "button"):
                
                // <button id="SaveImage" x="70" y="10" w="26" txt="Save Image" tooltip="Click to Save Image" />
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                txt = SafeXmlParameter(child, 'txt')
                tooltip = SafeXmlParameter(child, 'tooltip')   
                
                // CreateButton(frm:^forms, x:Integer, y:Integer, w:Integer, txt:String, tooltip:String="")
                guiButton = CreateButton(parentForm, x, y, w, txt, tooltip)
                guiButton.hook = hook
                    
                
            // ImageButton
            (s = "image-button"):
            
                // <image-button id="SaveImage" x="70" y="10" w="26" txt="Save Image" tooltip="Click to Save Image" file="foo.png" />
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                txt = SafeXmlParameter(child, 'txt')
                tooltip = SafeXmlParameter(child, 'tooltip')
                fn = SafeXmlParameter(child, 'file')
                f = x2g_stoi(SafeXmlParameter(child, 'frames'))
                If f < 1 then f = 1
                
                // CreateButton(frm:^forms, x:Integer, y:Integer, w:Integer, txt:String, tooltip:String="")
                guiButton = CreateButton(parentForm, x, y, w, txt, tooltip)
                guiButton.hook = hook
                
                If FileExists(fn, FALSE) then
                    tmpImg = LoadImage(fn)
                    ImageButton(guiButton, CreateImage(ImageWidth(tmpImg), ImageHeight(tmpImg), f))
                    
                    For k = 0 to (f-1)
                        ImageFrame(guiButton.image, k)
                        DrawImage(tmpImg, 0, 0, guiButton.image)
                    Next
                    FreeImage(tmpImg)
                    
                Endif
                ImageFrame(guiButton.image, 0)
               
             
            // Canvas
            (s = "canvas"):
            
                // <canvas id="mainCanvas" x="10" y="10" w="10" h="10" tooltip="" [file="foo.png"] />
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                h = x2g_stoi(SafeXmlParameter(child, 'h'))
                tooltip = SafeXmlParameter(child, 'tooltip')
                fn = SafeXmlParameter(child, 'file')
                
                // CreateCanvas(frm:^forms, x:Integer, y:Integer, w:Integer, h:Integer, tooltip:String="")
                guiCanvas = CreateCanvas(parentForm, x, y, w, h, tooltip)
                guiCanvas.hook = hook
                
                If FileExists(fn, FALSE) then
                    tmpImg = LoadImage(fn)
                    CLS(ToRGBA(0,0,0,0), guiCanvas.e)
                    DrawImage(tmpImg, 0, 0, guiCanvas.e)
                    FreeImage(tmpImg)
                Endif
                
            
            // Checkbox
            (s = 'check-box'):

                // <check-box x="12" y="72" w="120" checked="TRUE|FALSE" txt="Enabled" tooltip="Click to toggle" />
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                txt = SafeXmlParameter(child, 'txt')
                tooltip = SafeXmlParameter(child, 'tooltip')
                checked = (Upper(SafeXmlParameter(child, 'checked')) = "TRUE")
            
                // CreateCheckBox(frm: ^forms, x:Integer, y:Integer, w:Integer, txt:String, checked:Boolean=FALSE, tooltip:String="")
                guiCheckBox = CreateCheckBox(parentForm, x, y, w, txt, checked, tooltip)
                guiCheckBox.hook = hook
                
            
            // DropDown
            (s = 'drop-down'):
            
                // <drop-down x="12" y="72" w="120" txt="Preview" tooltip="Select Editor Mode"> ... </drop-down>
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                txt = SafeXmlParameter(child, 'txt')
                tooltip = SafeXmlParameter(child, 'tooltip')
            
                // CreateDropDown(frm: ^forms, x:Integer, y:Integer, w:Integer, txt:String, tooltip:String="")
                guiDropDown = CreateDropDown(parentForm, x, y, w, txt, tooltip)
                guiDropDown.hook = hook
                
                // Get child options
                c2 = PoxChildren(child)
                
                For j = 0 to (c2-1)
                    optionChild = PoxChild(child, j)
                    s = optionChild.name.s
                    
                    If (s = 'option') then
                        txt = SafeXmlParameter(optionChild, 'txt')
                        DropDownAddOption(guiDropDown, txt)
                    Else
                        Xml2Gui_error("Element '"+s+"' is not an option. Only option elements may appear inside drop-downs.", fn)
                    Endif
                Next
            
            
            // Frame
            (s = 'frame'):
            
                // <frame id="tabs1" style="tab" x="10" y="10" w="10" h="10" title="My Frame")
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                h = x2g_stoi(SafeXmlParameter(child, 'h'))
                style = Upper(SafeXmlParameter(child, 'style'))
                title = SafeXmlParameter(child, 'title')
                            
                Conditions
                    (style = "TAB"):            styleId = 0
                    (style = "ROUNDBORDER"):    styleId = 1
                    (style = "ROUNDSOLID"):     styleId = 2
                    (style = "SQUAREBORDER"):   styleId = 3
                    (style = "SQUARESOLID"):    styleId = 4
                    (style = "MOCKWINDOW"):     styleId = 5
                    (style = "HEADER"):         styleId = 6
                
                    Default:
                        Xml2Gui_error("Unkown tab style '"+style+"'.", fn)
                        styleId = 4
                EndConditions
            
                // CreateFrame(frm:^forms, x:Integer, y:Integer, w:Integer, h:Integer, style:Integer, title:String="")
                guiFrame = CreateFrame(parentForm, x, y, w, h, styleId, title)
                guiFrame.hook = hook
            
            
            // List Box
            (s = "list-box"):

                // <list-box id="listbox1" x="10" y="10" w="100" h="300" tooltip="Select item"> ... </list-box>
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                h = x2g_stoi(SafeXmlParameter(child, 'h'))
                tooltip = SafeXmlParameter(child, 'tooltip')
            
                // CreateListBox(frm: ^forms, x:Integer, y:Integer, w:Integer, h:Integer, tooltip:String="")
                guiListBox = CreateListBox(parentForm, x, y, w, h, tooltip)
                guiListBox.hook = hook
                
                // Get child options
                c2 = PoxChildren(child)
                
                For j = 0 to (c2-1)
                    optionChild = PoxChild(child, j)
                    s = optionChild.name.s
                    
                    If (s = 'item') then
                        txt = SafeXmlParameter(optionChild, 'txt')
                        ListBoxAddOption(guiListBox, txt)
                    Else
                        Xml2Gui_error("Element '"+s+"' is not an item. Only item elements may appear inside list-boxes.", fn)
                    Endif
                Next
            
            
            // Progress Bar
            (s = 'progress-bar'):
                
                // <progressBar id="complete" x="10" y="10" l="100" o="0|1" tooltip="Percent of file loaded") 
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                l = x2g_stoi(SafeXmlParameter(child, 'l'))
                o = x2g_stoi(SafeXmlParameter(child, 'o'))
                p = x2g_stoi(SafeXmlParameter(child, 'percent')) 
                tooltip = SafeXmlParameter(child, 'tooltip')
            
                //  CreateProgressBar(frm:^forms, x:Integer, y:Integer, l:Integer, ori:Integer=HAG_HORIZONTAL, tooltip:String="")
                guiProgressBar = CreateProgressBar(parentForm, x, y, l, o, tooltip)
                ProgressBarPercent(guiProgressBar, p)
                guiProgressBar.hook = hook
            
            
            
            // Slider
            (s = "slider"):
                
                // <slider id="time" x="10" y="10" l="100" d="60" o="0|1" tooltip="Time passed") 
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                l = x2g_stoi(SafeXmlParameter(child, 'l'))
                d = x2g_stoi(SafeXmlParameter(child, 'd'))
                o = x2g_stoi(SafeXmlParameter(child, 'o'))
                tooltip = SafeXmlParameter(child, 'tooltip')
            
                // CreateSlider(frm:^forms, x:Integer, y:Integer, l:Integer, divisions: Integer = 0, ori:Integer=HAG_HORIZONTAL, tooltip:String="")
                guiSlider = CreateSlider(parentForm, x, y, l, d, o, tooltip)
                guiSlider.hook = hook
                
            
            // TabGroup
            (s = "tab-group"):
                
                // <tab-group id="openFilesTab"> ... </tab-group>
                guiTabGroup = CreateTabGroup(parentForm)
                guiTabGroup.hook = hook
                
                // Get child tabs
                c2 = PoxChildren(child)
                
                For j = 0 to (c2-1)
                    tabChild = PoxChild(child, j)
                    s = tabChild.name.s
                    hook = prefix+"."+SafeXmlParameter(tabChild, 'id')
                    
                    If (s = 'tab') then
                    
                        // <tab id="id" x="70" y="10" w="26" selected="TRUE|FALSE" txt="foo" tooltip="foo" />
                        x = x2g_stoi(SafeXmlParameter(tabChild, 'x'))
                        y = x2g_stoi(SafeXmlParameter(tabChild, 'y'))
                        w = x2g_stoi(SafeXmlParameter(tabChild, 'w'))
                        txt = SafeXmlParameter(tabChild, 'txt')
                        tooltip = SafeXmlParameter(tabChild, 'tooltip')
                        selected = (Upper(SafeXmlParameter(tabChild, 'selected')) = "TRUE")
                        
                        // CreateTab(grp: ^tabGroups, x:Integer, y:Integer, w:Integer, txt:String, selected:Boolean=FALSE, tooltip:String="")
                        guiTab = CreateTab(guiTabGroup, x, y, w, txt, selected, tooltip)
                        guiTab.hook = hook
                        
                    Else
                        Xml2Gui_error("Element '"+s+"' is not a tab. Only tab elements may appear inside tab groups.", fn)
                    Endif
                Next
            
            
            // Text
            (s = "text"):

                // <text x="600" y="48" w="90" txt="Mode:" />
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                h = x2g_stoi(SafeXmlParameter(child, 'h'))
                txt = SafeXmlParameter(child, 'txt')
                tooltip = SafeXmlParameter(child, 'tooltip')
                styleId = x2g_stoi(SafeXmlParameter(child, 'style'))
                If w = 0 then w = -1
                If h = 0 then h = -1 
                
                // CreateText(frm: ^forms, x:Integer, y:Integer, txt:String, w:Integer=-1, h:Integer=-1, ttype:Integer=TXT_TYPE_NORMAL, tooltip:String="")
                guiText = CreateText(parentForm, x, y, txt, w, h, styleId, tooltip)
                guiText.hook = hook
                
                
            // Text Box
            (s = "text-box"):

                // <Text-box x="600" y="48" w="90" txt="Edit Me!" tooltip="I can be edited" />
                x = x2g_stoi(SafeXmlParameter(child, 'x'))
                y = x2g_stoi(SafeXmlParameter(child, 'y'))
                w = x2g_stoi(SafeXmlParameter(child, 'w'))
                txt = SafeXmlParameter(child, 'txt')
                tooltip = SafeXmlParameter(child, 'tooltip')
                If w = 0 then w = -1
                If h = 0 then h = -1 
                
                // CreateTextBox(frm: ^forms, x:Integer, y:Integer, w:Integer, txt:String, tooltip:String="")
                guiTextBox = CreateTextBox(parentForm, x, y, w, txt, tooltip)
                guiTextBox.hook = hook
            
            
            Default:
                Xml2Gui_error("Unkown element: '"+s+"'", fn)
                
            
        EndConditions
        
    Next

End

Procedure Xml2Gui(fn: String) ; export
Var
    rootForm: ^forms
    guiXml: ^poxFiles
    child: ^poxNodes
    root: ^poxNodes
    i, c: Integer
Begin
    
    guiXml = PoxLoad(fn)
    root = PoxRootNode(guiXml)
    c = PoxChildren(root)
    
    For i = 0 to (c-1)
        
        child = PoxChild(root, i)
        
        If child.name.s = 'form' then
            rootForm = CreateForm(SafeXmlParameter(child, 'name'))            
            Xml2Gui_LoadForm(rootForm, child, SafeXmlParameter(child, 'id'))
        Else
            Xml2Gui_error("Element '"+child.name.s+"' has no parent form.", fn)
        Endif
    
    Next
        
End

Function x2g_stoi(s: String): Integer
Begin
    If x2g_StringIsInt(s) then
        result = ToInt(s)
    Else
        result = 0
    Endif
End

Function x2g_CharIsNumeric(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)    
    If (a >= 48) and (a <= 57) then result = TRUE else result = FALSE    
End

Function x2g_StringIsInt(s: String) : Boolean
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (x2g_CharIsNumeric(c) or (c = '-')) then exit
        If (i <> 1) and (c = '-') then exit 
    Next
    
    result = TRUE        
End

Procedure Xml2Gui_error(s: String, fn: String)
Begin
    MessageBox("Error in '"+fn+"': "+s, "xml2Gui")
End

Begin
End
